#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Jan 26 16:31:50 2021

@author: lbresin
"""
 

import pandas as pd
import numpy as np

### Schmidt et al Soil Water balance model, Schmidt et al, 2014 (doi: 10.1016/j.jhydrol.2014.02.019) ####

def compute_eta(ETp, P, wilting_point, SMS_prev):
    '''
    This functions computes the actual Evapotranspiration (mm)
    
    Parameters
    ----------
    SMS_prev : Numpy Array
        Soil Moistore Storage at the previous timestep (mm).
    P : Numpy Array
        Precipitation (mm).
    wilting_point : Numpy Array
        Wilting Point (mm).
    ETp : Numpy Array
        Potential Evapotranspiration (mm).

    Returns
    -------
    ETa : Numpy Array
        Actual Evapotranspiration (mm).

    '''
    # SMS_prev + P - WP < ETp
    ETa = SMS_prev + P - wilting_point
    
    # SMS_prev + P - WP >= ETp
    ETa[np.where((SMS_prev + P - wilting_point)>=ETp)] = ETp[np.where((SMS_prev + P - wilting_point)>=ETp)]
    
    # SMS_prev + P < WP
    ETa[np.where(SMS_prev + P < wilting_point)] = 0
    
    return ETa

def compute_dp(ETa, P, SMS_prev, field_capacity):
    '''
    This function computes the deep percolation (mm)

    Parameters
    ----------
    ETa : Numpy Array
        Actual Evapotranspiration (mm).
    P : Numpy Array
        Precipitation (mm).
    SMS_prev : Numpy Array
        Soil Moistore Storage at the previous timestep (mm).
    field_capacity : Numpy Array
        Field Capacity (mm).

    Returns
    -------
    DP : Numpy Array
        Deep percolation (mm).

    '''
    # SMS_prev + P - ETa <= FC
    DP = np.zeros(ETa.shape)
    
    # SMS_prev + P - ETa > FC
    DP[np.where((SMS_prev + P - ETa) > field_capacity)] = (SMS_prev + P - ETa - field_capacity)[np.where((SMS_prev + P - ETa) > field_capacity)]
    
    return DP

def compute_sm(P, ETa, DP, SMS_prev):
    '''
    This function compute the Soil moisture at t_i (mm)

    Parameters
    ----------
    P : Numpy Array
        Precipitation (mm).
    ETa : Numpy Array
        Actual Evapotranspiration (mm).
    DP : Numpy Array
        Deep percolation (mm).
    SMS_prev : Numpy Array
        Soil Moistore Storage at the previous timestep (mm).

    Returns
    -------
    SMS_curr : Numpy Array
        Soil Moistore Storage at the current timestep (mm).

    '''
    SMS_curr = SMS_prev + P - ETa - DP
    
    return SMS_curr
    
def initialize_m1(P, ETp, SMS_initial, wilting_point, field_capacity):
    '''
    This functions intializes the first timestep with an initial soil moistore storage.

    Parameters
    ----------
    P : Numpy Array
        Precipitation (mm).
    ETp : Numpy Array
        Potential Evapotranspiration (mm).
    SMS_initial : Numpy Array
        Initial Soil Moistore Storage (mm).
    wilting_point : Numpy Array
        Wilting Point (mm).
    field_capacity : Numpy Array
        Field Capacity (mm).

    Returns
    -------
    stor_curr : Dictionary of Numpy Arrays
        A dictionary with the variables ETall computed and input variables of the current timestep.

    '''
    
    stor_curr = dict()

    stor_curr['ETa'] = compute_eta(ETp, P, wilting_point, SMS_initial) 
    stor_curr['DP'] = compute_dp(stor_curr['ETa'], P, SMS_initial, field_capacity)
    stor_curr['SMS']  = compute_sm(P, stor_curr['ETa'], stor_curr['DP'], SMS_initial)
    
    return stor_curr
  
def iterate_through_time_m1(P, ETp, wilting_point, field_capacity, stor_curr):
    '''
    Iterate through time

    Parameters
    ----------
    P : Numpy Array
        Precipitation (mm).
    ETp : Numpy Array
        Potential Evapotranspiration (mm).
    wilting_point : Numpy Array
        Wilting Point (mm).
    field_capacity : Numpy Array
        Field Capacity (mm).
    stor_curr : Dictionary of Numpy Arrays
        A dictionary with the variables ETall computed and input variables of the current timestep.

    Returns
    -------
    stor_curr : Dictionary of Numpy Arrays
        A dictionary with the variables ETall computed and input variables of the current timestep.

    '''
    stor_prev = stor_curr
    
    stor_curr['ETa'] = compute_eta(ETp, P, wilting_point, stor_prev['SMS']) 
    stor_curr['DP'] = compute_dp(stor_curr['ETa'], P, stor_prev['SMS'], field_capacity)
    stor_curr['SMS']  = compute_sm(P, stor_curr['ETa'], stor_curr['DP'], stor_prev['SMS'])
       
    return stor_curr

def initialize_m2(P, ETp, SMS_initial, wilting_point, field_capacity, m1_fraction):
    '''
    This functions intializes the first timestep of a dual medium approach with an initial soil moistore storage.

    Parameters
    ----------
    P : Numpy Array
        Precipitation (mm).
    ETp : Numpy Array
        Potential Evapotranspiration (mm).
    SMS_initial : Dictionary of Numpy Arrays
        Initial Soil Moistore Storage (mm).
    wilting_point : Dictionary of Numpy Arrays
        Wilting Point (mm).
    field_capacity : Dictionary of Numpy Arrays
        Field Capacity (mm).
    m1_fraction : Float
        Areal Fraction of medium 1
        
    Returns
    -------
    stor_curr : Dictionary of Numpy Arrays
        A dictionary with the variables ETall computed and input variables of the current timestep.

    '''
    
    stor_curr = dict()
    
    stor_curr['ETa_1'] = compute_eta(ETp, P, wilting_point['1'], SMS_initial['1']) 
    stor_curr['DP_1'] = compute_dp(stor_curr['ETa_1'], P, SMS_initial['1'], field_capacity['1'])
    stor_curr['SMS_1']  = compute_sm(P, stor_curr['ETa_1'], stor_curr['DP_1'], SMS_initial['1'])
    
    stor_curr['ETa_2'] = compute_eta(ETp, P, wilting_point['2'], SMS_initial['2']) 
    stor_curr['DP_2'] = compute_dp(stor_curr['ETa_2'], P, SMS_initial['2'], field_capacity['2'])
    stor_curr['SMS_2']  = compute_sm(P, stor_curr['ETa_2'], stor_curr['DP_2'], SMS_initial['2'])   
    
    stor_curr['DP_total']  = stor_curr['DP_1']*m1_fraction + stor_curr['DP_2']*(1-m1_fraction)
    
    return stor_curr
  
def iterate_through_time_m2(P, ETp, wilting_point, field_capacity, stor_curr, m1_fraction):
    '''
    Iterate through time of a dual medium approach 

    Parameters
    ----------
    P : Numpy Array
        Precipitation (mm).
    ETp : Numpy Array
        Potential Evapotranspiration (mm).
    wilting_point : Numpy Array
        Wilting Point (mm).
    field_capacity : Numpy Array
        Field Capacity (mm).
    stor_curr : Dictionary of Numpy Arrays
        A dictionary with the variables ETall computed and input variables of the current timestep.
    m1_fraction : Float
        Areal Fraction of medium 1
        
    Returns
    -------
    stor_curr : Dictionary of Numpy Arrays
        A dictionary with the variables ETall computed and input variables of the current timestep.

    '''
    stor_prev = stor_curr
    
    stor_curr['ETa_1'] = compute_eta(ETp, P, wilting_point['1'], stor_prev['SMS_1']) 
    stor_curr['DP_1'] = compute_dp(stor_curr['ETa_1'], P, stor_prev['SMS_1'], field_capacity['1'])
    stor_curr['SMS_1']  = compute_sm(P, stor_curr['ETa_1'], stor_curr['DP_1'], stor_prev['SMS_1'])
    
    stor_curr['ETa_2'] = compute_eta(ETp, P, wilting_point['2'], stor_prev['SMS_2']) 
    stor_curr['DP_2'] = compute_dp(stor_curr['ETa_2'], P, stor_prev['SMS_2'], field_capacity['2'])
    stor_curr['SMS_2']  = compute_sm(P, stor_curr['ETa_2'], stor_curr['DP_2'], stor_prev['SMS_2'])   
    
    stor_curr['DP_total']  = stor_curr['DP_1']*m1_fraction + stor_curr['DP_2']*(1-m1_fraction)
       
    return stor_curr


if __name__ == "__main__":
    #### OPEN INPUT ####  
    df_P = pd.read_csv('./climate_model/P_mm_d-1.csv', index_col=0) #mm/d
    #df_P = df_P.set_index('datetime')
    df_P.columns = df_P.columns.astype(int)
    df_P.index = [pd.Timestamp(i) for i in df_P.index]
    df_P = df_P.reindex(sorted(df_P.columns), axis=1)
    #df_P *= 10*100 # convert to mm/d
    
    df_ETp = pd.read_csv('./climate_model/ETp_mm_d-1.csv', index_col=0) #mm/d
    #df_ETp = df_ETp.set_index('datetime')
    df_ETp.columns = df_ETp.columns.astype(int)
    df_ETp.index = [pd.Timestamp(i) for i in df_ETp.index]
    df_ETp = df_ETp.reindex(df_P.columns, axis=1)
    #df_ETp *= 10*100 # convert to mm/d
  
    # get rid of non-overlapping Time
    date_start = max([df_P.index.min(), df_ETp.index.min()])
    date_end = min([df_P.index.max(), df_ETp.index.max()])
    df_P = df_P[(df_P.index>=date_start) & (df_P.index<=date_end)]
    df_ETp = df_ETp[(df_ETp.index>=date_start) & (df_ETp.index<=date_end)]
      
    ### COMPUTE DEEP PERCOLATION VIA DUAL MEDIUM MODEL, according to Schmidt et al, 2014 (doi: 10.1016/j.jhydrol.2014.02.019) ####
    stor_prev, stor_curr, stor_DP = None, None, None # reset so the programm knows it need to initialze 
    df_DP_schmidt_dual = pd.DataFrame(columns=df_P.columns)
    
    for date in df_P.index:
        P = df_P.loc[date,:].to_numpy()
        ETp = df_ETp.loc[date,:].to_numpy()    
        
        if stor_curr == None: # The model parameters (i.e. initial soil moisture) needs to be initialzed in the first iteration        
            df_schmidt_param_dual = pd.read_csv('./parameter_dual.csv', index_col=0)
            
            SMS_initial = {'1':df_schmidt_param_dual.loc['initial_soil_moisture_1',:].to_numpy(),
                           '2':df_schmidt_param_dual.loc['initial_soil_moisture_2',:].to_numpy()} # mm
            field_capacity = {'1': df_schmidt_param_dual.loc['field_capacity_1',:].to_numpy(),
                              '2': df_schmidt_param_dual.loc['field_capacity_2',:].to_numpy()}# mm
            wilting_point = {'1': df_schmidt_param_dual.loc['wilting_point_1',:].to_numpy(),
                             '2': df_schmidt_param_dual.loc['wilting_point_2',:].to_numpy()}# mm
            m1_fraction = df_schmidt_param_dual.loc['fraction_1',:].to_numpy()

            # Initialize Model
            stor_curr = initialize_m2(P, ETp, SMS_initial, wilting_point, field_capacity, m1_fraction)
            
        else: # All subsequent iterations
            stor_curr = iterate_through_time_m2(P, ETp, wilting_point, field_capacity, stor_curr, m1_fraction)
        
        df_DP_schmidt_dual.loc[date, :] = stor_curr['DP_total']
        
    df_DP_schmidt_dual_si_unit = df_DP_schmidt_dual/10/100
        
    df_DP_schmidt_dual.to_csv('./lbresin_dp_climate_model_1981-2070_mm_d-1.csv')
    df_DP_schmidt_dual_si_unit.to_csv('./lbresin_dp_climate_model_1981-2070_m_d-1.csv')